/*
 *  Linux kernel adaptation layers for owl wifi library.
 *  Copyright (C) 2010 HD Wireless AB
 *
 */

#include <linux/version.h>

#include "owl_raw.h"
#include "owl_net.h"
#include "owl_drv.h"

struct owl_raw {
        struct sk_buff_head rxq;
        struct net_device *dev;
        struct owl_transport *transport;
};

/* Yuck */
#define HIC_MESSAGE_TYPE_CTRL        7
#define MAC_API_PRIMITIVE_TYPE_RSP    0x80
#define MAC_API_PRIMITIVE_TYPE_CFM    0x80
#define HIC_CTRL_SET_ALIGNMENT_CFM       (4  | MAC_API_PRIMITIVE_TYPE_CFM)
#define HIC_CTRL_INIT_COMPLETED_CFM      (8  | MAC_API_PRIMITIVE_TYPE_CFM)

/*
 * Invoked from owl_drv.c work queue when an incoming packet should be
 * processed. Will put all data except aligncfm i a queue that is read
 * through the RAWRX ioctl.
 *
 */
void owl_raw_rx(struct owl_transport *transport, struct sk_buff *skb)
{
        struct owl_raw *priv = transport->net;
        struct net_device *dev = priv->dev;
        int rx_len;
        char *rx_buf = skb->data;
        
        dev_dbg(&dev->dev, "owl_raw_rx got pkt of size %d\n", skb->len);
#ifdef VERBOSE_DEBUG
        print_hex_dump_bytes("raw RX ", DUMP_PREFIX_NONE,
                             skb->data, skb->len);
#endif
        rx_len = skb->len;
        if (rx_len > 4 && 
            rx_buf[2] == (HIC_MESSAGE_TYPE_CTRL | MAC_API_PRIMITIVE_TYPE_RSP) &&
            rx_buf[3] == (HIC_CTRL_SET_ALIGNMENT_CFM)) {
            
                /* Init complete request */
                const uint8_t req[] = {
                        0x0E, 0x00, 0x07, 0x07,
                        0x08,
                        0xFF,
                        0x00, 0x00,
                        0x02, 0x00, 0x00, 0x00,
                        0x00, 0x00, 0x00, 0x00
                };
                struct sk_buff *tx_skb;
                struct owl_raw *priv; 

                dev_dbg(&dev->dev, "Dropping align cfm\n");
                       
                priv = netdev_priv(dev);
                if ((tx_skb = 
                     dev_alloc_skb(sizeof(req))) == NULL)
                        goto out;
                        
                memcpy(skb_put(tx_skb, sizeof(req)), 
                       req, 
                       sizeof(req));
                priv->transport->tx(priv->transport, tx_skb);
                        
                return;
        }
        
        /* We sent the init complete request so we should not forward the
           reply */
        if (rx_len > 4 && 
            rx_buf[2] == (HIC_MESSAGE_TYPE_CTRL | MAC_API_PRIMITIVE_TYPE_RSP) &&
            rx_buf[3] == (HIC_CTRL_INIT_COMPLETED_CFM)) {
                
                dev_dbg(&dev->dev, "Dropping init completed cfm\n");
                return;
        }
out:        
        skb_queue_tail(&priv->rxq, skb);
}


static int owl_raw_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
        int err;
        struct ioc_owl_raw ioc;
        struct sk_buff *skb;
        struct owl_raw *priv = netdev_priv(dev);

        dev_dbg(&dev->dev, "cmd:%d\n", cmd);
        
        if ((err = copy_from_user(&ioc, ifr->ifr_data, sizeof(ioc)))) {
                dev_err(&dev->dev, "copy_from_user failed\n");                
                return -EFAULT;
        }
        
        switch (cmd) {
        case SIOCOWLRAWTX:
                if ((skb = dev_alloc_skb(ioc.len)) == NULL) {
                        dev_err(&dev->dev, "coult not alloc skb\n");
                        return -ENOMEM;
                }
        
                memcpy(skb_put(skb, ioc.len), ioc.data, ioc.len);
                priv->transport->tx(priv->transport, skb);
                return 0;
                
        case SIOCOWLRAWRX:
                if ((skb = skb_dequeue(&priv->rxq)) == NULL) {
                        ioc.len = 0;
                        goto out;
                }

                if (skb->len > sizeof(ioc.data)) {
                        dev_err(&dev->dev, "invalid skb len\n");
                        return -EFAULT;
                }
                
                ioc.len = skb->len;
                memcpy(ioc.data, skb->data, skb->len);
                dev_kfree_skb(skb);
        out:
                err = copy_to_user(ifr->ifr_data, &ioc, sizeof(ioc));
                if(err) {
                        dev_err(&dev->dev, "copy_to_user failed\n");
                        return -EFAULT;
                }

                return 0;

        default:
                dev_err(&dev->dev, "invalid ioctl\n");
                return -EFAULT;
        };
}
        
static int dummy_tx(struct sk_buff *skb, struct net_device *dev)
{
        return NETDEV_TX_OK;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
static const struct net_device_ops netdev_ops = {
        .ndo_do_ioctl = owl_raw_ioctl,
        .ndo_start_xmit         = dummy_tx,
};
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29) */

/*
 * Called by sdio driver when the card is inserted and initialized.
 *
 * Will initialize the network interface. Note that this call will update the
 * net field in @transport.
 *
 */
int owl_raw_register_drv(struct owl_transport *transport)
{
        int err;
        struct owl_raw *priv;
        struct net_device *dev = NULL;
        struct sk_buff *skb;

/* Valid for 0_0_12_74 + fw 6.12.17.4 */
        const uint8_t align_req[] = {0x1E, 0x00, 0x07, 0x03, 
                                     0x0A,       /* hic hdr size */
                                     0x46,       /* rsv */
                                     0x04, 0x00, /* pad bytes */
                                     0x20, 0x5B, /* hdr padding */
                                     0x01, 0x00, 0x00, 0x00, /* transid */
                                     0x20, 0x00, /* min size */
                                     0x10, 0x00, /* pad size */
                                     0x01,       /* host attention */
                                     0x10,       /* ul header size */
                                     0x00,       /* swap */
                                     0xFF,       /* hWakeup */
                                     0xFF,       /* hForceInterval */
                                     0x05,       /* tx_window_size */
                                     0x00, 0x00, /* block_size */
                                     0xFF, 0xFF, /* reserved */
                                     0x55, 0x53, 0x46, 0x52 /* message pad */
        };
        BUG_ON(!transport);
        BUG_ON(!transport->tx);
        BUG_ON(!transport->fw_download);
        BUG_ON(transport->net);

        if ((dev = alloc_etherdev(sizeof(*priv))) == NULL)
                return -ENOMEM;

        priv = netdev_priv(dev);
        memset(priv, 0, sizeof(*priv));
        transport->net = priv;
        priv->dev = dev;
        priv->transport = transport;
        skb_queue_head_init(&priv->rxq);        

        if ((err = dev_alloc_name(dev, "owl%d")) < 0) {
                dev_err(&dev->dev, "dev_alloc_name(), err=%d\n", err); 
                goto err;
        }

        /* setup netdev ops */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
        dev->netdev_ops = &netdev_ops;
#else
        dev->do_ioctl = owl_raw_ioctl;
#endif
        if ((err = register_netdev(dev))) {
                dev_err(&dev->dev, "register_netdev(), err=%d\n", err);
                goto err;
        }

        if ((err = transport->fw_download(transport))) {
                dev_err(&dev->dev, "fw download failed, err=%d\n", err);
                return err;
        }

        netif_carrier_off(dev);

        if ((skb = dev_alloc_skb(sizeof(align_req))) == NULL)
                goto err;
        
	memcpy(skb_put(skb, sizeof(align_req)), 
               align_req, 
               sizeof(align_req));
        transport->tx(transport, skb);

        dev_info(&dev->dev, "ready\n");
        return 0;

err:
        dev_info(&dev->dev, "error\n");
        free_netdev(dev);
        return err;
}


/*
 * Called by sdio driver when the card is removed or if the driver is unloaded.
 *
 */
void owl_raw_unregister_drv(struct owl_transport *transport)
{
        struct owl_raw *priv = transport->net;
        struct sk_buff *skb;
        struct net_device *dev;

        BUG_ON(!priv);
        BUG_ON(!priv->dev);
        
        dev = priv->dev;
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);

        while ((skb = skb_dequeue(&priv->rxq)))
                dev_kfree_skb_any(skb); /* if in irq context */

        unregister_netdev(priv->dev);

        /* will also free priv */
        free_netdev(priv->dev);

        dev_info(&dev->dev, "shutdown\n");
}
