/*
 *  Programming interface for wlap_api.
 *  Copyright (C) 2011 HD Wireless AB
 *
 *  You should have received a copy of the license along with this library.
 */

/*! \file wlap_api.h *************************************************************
 *
 * \brief WiFi AP API
 *
 * This file provides the wlap_api interface.
 *
 * - Compiler:           GNU GCC for AVR32
 * - Supported devices:  
 *  \li SPB104 + EVK1100
 *  \li SPB104 + EVK1101
 *  \li SPB104 + EVK1104
 *  \li SPB104 + EVK1105 (SPI)
 *  \li SPB104 + EVK1105 (SPI + irq)
 *  \li SPB105 + EVK1105 (SPI)
 * - AppNote:
 *
 * \author               H&D Wireless AB: \n
 *
 *****************************************************************************
 *
 * \section intro Introduction
 * This is the documentation for the WiFi AP Driver API \a wlap_api.
 *
 * \section files Main Files
 * - wlap_api.h : WiFi driver interface.
 * - libwlap_api_*.*.a - Driver library.
 * 
 */

#ifndef WLAP_API_H
#define WLAP_API_H
 
#define WLAP_API_RELEASE_NAME "unknown"

#include <wl_api.h>

/** \defgroup wl_softap Access Point Mode
 *
 * \brief Support the WiFi Access Point mode.
 *
 *  @{
 */
/*
 * Station representation 
 *
 */
struct wl_sta_t
{
        struct wl_mac_addr_t bssid; /**<  The BSSID of the network. */
        uint8_t queued_pkt_cnt;     /**< Number of queueud packets for 
                                       this STA. */
        uint8_t in_ps;              /**< Is the STA in power save mode. */
        uint8_t aid;                /**< STA AID */
};

/* Station list representation. Array of pointers to wl_sta_t entries. */
struct wl_sta_list_t
{
        struct wl_sta_t **sta; /**< The list of pointers to stations */
        size_t cnt;            /**< Number of stations               */
};

/*! \brief Get the list of currently associated stations (SoftAP).
 *
 * Retrieves the list of current stations from 
 * the driver. 
 *
 * This function is not thread safe. It must be called in the
 * same execution context as wl_poll().
 *
 * @param network_list Output buffer. The API call returns
 *        a pointer to allocated memory containing the network list.
 * @return
 * - WL_SUCCESS
 * - WL_FAILURE.
 */
wl_err_t wlap_get_sta_list(struct wl_sta_list_t **network_list);


/*! Callback used to read data from a TX packet.
 *
 * This function is supplied by the user of the API (that means you).
 *
 * @param dst Destination buffer. The data should be copied
 *  to this buffer.
 * @param src_handle Handle to the source packet from where
 *  the data should be copied. This handle is the same one that
 *  is passed in parameter \a pkt_handle to \a wl_process_tx().
 * @param read_len Number of bytes to copy from \a src_handle
 *  to \a dst.
 * @param offset The offset in bytes, counting from the
 *  beginning of the Ethernet header, from where to copy data.
 * @return
 *  - The number of bytes copied. This number may be smaller
 *    than the length requested in \a read_len but it may not
 *    be shorter than the length of the packet counting from
 *    \a offset. In other words, if the caller of this function
 *    receives a return count that is shorter than \a read_len
 *    he will assume that all packet data has been read.
 * - < 0 on error.
 */
typedef ssize_t (*wl_pkt_read_cb_t)(char *dst,
                                    void *src_handle,
                                    size_t read_len,
                                    int offset);

/*! \brief Register a data access function for TX packets (SoftAP).
 *
 * When a TX data packet has a different representation than a single
 * contiguous buffer in memory then a packet read function must be
 * implemented and registered with this call. Whenever the library
 * needs to read packet data it will call this function to do it.
 *
 * This function can be ignored if the TX packet representation is
 * a single contiguous buffer. This function is only needed in SoftAP
 * mode.
 *
 * @param pkt_read_cb Read callback. 
 * @param ctx Context 
 */
void wl_register_pkt_read_cb(wl_pkt_read_cb_t pkt_read_cb);

/*! \brief Start a network using the SoftAP mode.
 *
 * This call will cause the WiFi chip to start sending beacons
 * and accept associations from WiFi stations.
 *
 * If \a ssid_len is 0 then any currently announced network is
 * stopped.
 *
 */
wl_err_t wlap_start_ap(const char *ssid,
                       const size_t ssid_len,
                       const uint8_t channel);

/*! \brief Set the state of a stations 802.1x port.
 *
 * When the 802.1x port is closed only EAPOL traffic will be
 * accepted by the driver (other packets are discarded).  When the
 * 802.1x port is opened (for instance by a WPA authenticator that has
 * successfully configured encryption keys for the station)
 * all traffic is accepted.
 * @param bssid Address of the station whose 802.1x port should be set.
 * @param open Flag determining if the port should be opened (1) or
 *        closed (0).
 * @return WL_SUCCESS
 */
wl_err_t wlap_set_8021x_port(struct wl_mac_addr_t bssid, int open);

/*! \brief Disconnect a STA (SoftAP)
 * 
 * @param bssid The BSSID of the station to disconnect.
 * @param reason The reason code to be sent to the STA.
 * @return 
 *  - WL_SUCCESS
 *  - WL_FAILURE.
 */
wl_err_t wlap_disconnect_sta(const struct wl_mac_addr_t bssid,
                             int reason);

/*! \brief Deauthenticate a STA (SoftAP)
 * 
 * @param bssid The BSSID of the station to deauthenticate.
 * @param reason The reason code to be sent to the STA.
 * @return 
 *  - WL_SUCCESS
 *  - WL_FAILURE.
 */
wl_err_t wlap_deauthenticate_sta(const struct wl_mac_addr_t bssid,
                                 int reason);

/*! \brief STA MIC failure callback
 *
 * This function receives TKIP MIC failure indications.
 * When a associated station generates a MIC failure this could indicate
 * that someone is trying to tamper with the connection and TKIP
 * countermeasures processing should be triggered. The countermeasures
 * processing is handled outside of the driver and this callback can
 * be used to detect that countermeasures should be initiated.
 *
 * @param src The MAC address of the station causing the failure.
 * @param unicast Flag indicating wether the MIC failure was caused
 *        by a unicast frame (1) or a multicast frame (0).
 * @param ctx Opaque context pointer.
 */
typedef void (*wlap_mic_failure_cb_t)(struct wl_mac_addr_t src,
                                      int unicast,
                                      void *ctx);

/* ! \brief Register MIC failure callback.
 *
 * Register the function that should be called if a station MIC
 * failure is detected. This will disable any previously registered
 * MIC failure callback which will no longer be called.
 *
 * @param cb Callback to register. Can be NULL to deregister the
 *           currently registered callback.
 * @param ctx Opaque context pointer that will be passed back to the
 *            callback.
 * @return WL_SUCCESS.
 */
wl_err_t wlap_register_mic_failure_cb(wlap_mic_failure_cb_t cb,
                                      void *ctx);

/*! \brief EAPOL frame delivery callback
 *
 * This function receives EAPOL frames that comes in from stations.
 * The EAPOL frame will also be delivered to consumers via the
 * wl_process_rx() function (normally to the IP stack) so this
 * is an alternate way to get the EAPOL frames (useful if a 
 * external authenticator does not want to interfere with the
 * normal network interface).
 * 
 * @param pkt Pointer to the EAPOL packet, starting with the
 *            ethernet header. The packet will be freed
 *            when the call to this function returns.
 * @param pkt_len Length of the EAPOL packet.
 * @param ctx Opaque context pointer.
 */
typedef void (*wlap_eapol_frame_rx_cb_t)(const char *pkt,
                                         const size_t pkt_len,
                                         void *ctx);

/*! \brief Register EAPOL receive callback.
 *
 * Register the function that should be called if EAPOL frame is
 * received from a station. This will disable any previously
 * registered EAPOL receive callback which will no longer be called.
 *
 * @param cb Callback to register. Can be NULL to deregister the
 *           currently registered callback.
 * @param ctx Opaque context pointer that will be passed back to the
 *            callback.
 * @return WL_SUCCESS.
 */
wl_err_t wlap_register_eapol_frame_rx_cb(wlap_eapol_frame_rx_cb_t cb,
                                         void *ctx);

enum wlap_assoc_event_t {
        WL_EVENT_STA_ASSOC = 1,
        WL_EVENT_STA_AUTH,
        WL_EVENT_STA_DEAUTH,
        WL_EVENT_STA_DISASSOC
};

/*! \brief STA association event callback.
 *
 * This function gets calld whenever a station is associated,
 * disassociated or deauthenticated.
 *
 * @param event The type of association event.
 * @param addr The MAC address of the station.
 * @param reason_code A reason code for the event (dependent on
 *        the event type, this code comes from the MLME frame).
 *        This is set to -1 if the reason code in unused.
 * @param ies Buffer containing any information elements associated
 *        with the event. For WL_EVENT_STA_ASSOC it contains the
 *        information elements received in the association request
 *        from the station.
 * @param ielen Length of the \a ies buffer in bytes.
 * @param ctx Opaque context pointer.
 */
typedef void (*wlap_assoc_event_cb_t)(const enum wlap_assoc_event_t event,
                                      const struct wl_mac_addr_t addr,
                                      const int reason_code,
                                      const char *ies,
                                      const size_t ielen,
                                      void *ctx);

/*! \brief Register STA association event callback.
 *
 * Register the function that should be called when a station
 * association event occurs. Association event are
 * - A station has authenticated with us
 * - A station has associated with us
 * - A station has disassociated from us
 * - A station has deauthenticated from us
 *
 * This will disable any previously registered STA association event
 * callback which will no longer be called.
 *
 * @param cb Callback to register. Can be NULL to deregister the
 *           currently registered callback.
 * @param ctx Opaque context pointer that will be passed back to the
 *            callback.
 * @return WL_SUCCESS.
 */
wl_err_t wlap_register_assoc_event_cb(wlap_assoc_event_cb_t cb,
                                      void *ctx);

struct wl_receive_seq_cnt_t {
        uint8_t octet[8];
};

/* \brief Set an encryption key for a station.
 *
 */
wl_err_t wlap_set_key(uint8_t key_idx, 
                      size_t key_len, 
                      const void *key,
                      int swap,
                      const struct wl_mac_addr_t *bssid,
                      const struct wl_receive_seq_cnt_t *rsc,
                      int tx_key);

/*! \brief Set beacon paramaters for the AP.
 *

 * This function configures the beacon that is used in the AP mode.
 * 
 * @param head Buffer containing the beacon starting with the 802.11 beacon frame
 *        header and ending with the IE immediately preceding the TIM IE.
 * @param head_len Length of the \a head buffer in bytes.
 * @param tail Buffer containing the remainder of the beacon immediately following
 *        the TIM IE.
 * @param tail_len Length of the \a tail buffer in bytes.
 * @param dtim_period DTIM period to announce in beacons.
 * @param beacon_interval Interval with which to send beacons. 
 * @return WL_SUCCESS or WL_FAILURE.
 */
wl_err_t wlap_set_beacon(const uint8_t *head,
                         size_t head_len,
                         const uint8_t *tail,
                         size_t tail_len,
                         int dtim_period,
                         int beacon_interval);

/*! \brief Return the rates supported by the radio.
 *
 * @param rates An array of rates that will be filled
 *        in by this call. It must be at least \a len 
 *        elements long. A rate is counted in units
 *        of 500kbps so that the rate 11Mbps will
 *        have the representation 22 (as per the 802.11
 *        standards).
 * @param len On input this parameter should contain the
 *        number of elements in \a rates. On output it
 *        will contain the number of elements that were
 *        copied to \a rates (if the return value was
 *        WL_SUCCESS) or the number of elements required
 *        (if the return value was WL_INVALID_LENGTH).
 * @return 
 *        - WL_SUCCESS if the rates was successfully copied.
 *        - WL_INVALID_LENGTH if the \a rates parameter
 *          was too short to contain all the requested rates.
 */
wl_err_t wl_get_hw_rates(wl_rate_t *rates, size_t *len);

/*! @} */ /* End wl_softap group */


#endif
